from typing import Any, Dict, List

import cirq
import stim

@cirq.value_equality
class IIGate(cirq.Gate):
    """Handles explaining stim's II gate to cirq."""

    def _num_qubits_(self) -> int:
        return 2

    def _circuit_diagram_info_(self, args: cirq.CircuitDiagramInfoArgs) -> List[str]:
        return ['II', 'II']

    def _value_equality_values_(self):
        return ()

    def _decompose_(self, qubits):
        pass

    def _stim_conversion_(self, edit_circuit: stim.Circuit, targets: List[int], tag: str, **kwargs):
        edit_circuit.append('II', targets, tag=tag)

    def __pow__(self, power: int) -> 'IIGate':
        return self

    def __str__(self) -> str:
        return 'II'

    def __repr__(self):
        return f'stimcirq.IIGate()'

    @staticmethod
    def _json_namespace_() -> str:
        return ''

    def _json_dict_(self) -> Dict[str, Any]:
        return {}



import cirq
import stim
import stimcirq


def test_stim_conversion():
    a, b, c = cirq.LineQubit.range(3)

    cirq_circuit = cirq.Circuit(
        stimcirq.CXSwapGate(inverted=False).on(a, b),
        stimcirq.CXSwapGate(inverted=True).on(b, c),
    )
    stim_circuit = stim.Circuit(
        """
        CXSWAP 0 1
        TICK
        SWAPCX 1 2
        TICK
        """
    )
    assert stimcirq.cirq_circuit_to_stim_circuit(cirq_circuit) == stim_circuit
    assert stimcirq.stim_circuit_to_cirq_circuit(stim_circuit) == cirq_circuit


def test_diagram():
    a, b = cirq.LineQubit.range(2)
    cirq.testing.assert_has_diagram(
        cirq.Circuit(
            stimcirq.CXSwapGate(inverted=False)(a, b),
            stimcirq.CXSwapGate(inverted=True)(a, b),
        ),
        """
0: ---ZSWAP---XSWAP---
      |       |
1: ---XSWAP---ZSWAP---
        """,
        use_unicode_characters=False,
    )


def test_inverse():
    a = stimcirq.CXSwapGate(inverted=True)
    b = stimcirq.CXSwapGate(inverted=False)
    assert a**+1 == a
    assert a**-1 == b
    assert b**+1 == b
    assert b**-1 == a


def test_repr():
    val = stimcirq.CXSwapGate(inverted=True)
    assert eval(repr(val), {"stimcirq": stimcirq}) == val
    val = stimcirq.CXSwapGate(inverted=False)
    assert eval(repr(val), {"stimcirq": stimcirq}) == val


def test_equality():
    eq = cirq.testing.EqualsTester()
    eq.add_equality_group(stimcirq.CXSwapGate(inverted=False), stimcirq.CXSwapGate(inverted=False))
    eq.add_equality_group(stimcirq.CXSwapGate(inverted=True))


def test_json_serialization():
    a, b, d = cirq.LineQubit.range(3)
    c = cirq.Circuit(
        stimcirq.CXSwapGate(inverted=False)(a, b),
        stimcirq.CXSwapGate(inverted=False)(b, d),
    )
    json = cirq.to_json(c)
    c2 = cirq.read_json(json_text=json, resolvers=[*cirq.DEFAULT_RESOLVERS, stimcirq.JSON_RESOLVER])
    assert c == c2


def test_json_backwards_compat_exact():
    raw = stimcirq.CXSwapGate(inverted=True)
    packed = '{\n  "cirq_type": "CXSwapGate",\n  "inverted": true\n}'
    assert cirq.to_json(raw) == packed
    assert cirq.read_json(json_text=packed, resolvers=[*cirq.DEFAULT_RESOLVERS, stimcirq.JSON_RESOLVER]) == raw



import cirq
import stimcirq


def test_repr():
    r = stimcirq.IIErrorGate([0.25, 0.125])
    assert eval(repr(r), {'stimcirq': stimcirq}) == r
    r = stimcirq.IIErrorGate([])
    assert eval(repr(r), {'stimcirq': stimcirq}) == r


def test_json_serialization():
    r = stimcirq.IIErrorGate([0.25, 0.125])
    c = cirq.Circuit(r(cirq.LineQubit(1), cirq.LineQubit(3)))
    json = cirq.to_json(c)
    c2 = cirq.read_json(json_text=json, resolvers=[*cirq.DEFAULT_RESOLVERS, stimcirq.JSON_RESOLVER])
    assert c == c2


def test_json_backwards_compat_exact():
    raw = stimcirq.IIErrorGate([0.25, 0.125])
    packed = '{\n  "cirq_type": "IIErrorGate",\n  "gate_args": [\n    0.25,\n    0.125\n  ]\n}'
    assert cirq.read_json(json_text=packed, resolvers=[*cirq.DEFAULT_RESOLVERS, stimcirq.JSON_RESOLVER]) == raw
    assert cirq.to_json(raw) == packed



import functools
import itertools
import math
from typing import Callable, cast, Dict, Iterable, List, Optional, Sequence, Tuple, Type

import cirq
import stim

from ._i_error_gate import IErrorGate
from ._ii_error_gate import IIErrorGate
from ._ii_gate import IIGate


def _forward_single_str_tag(op: cirq.CircuitOperation) -> str:
    tags = [tag for tag in op.tags if isinstance(tag, str)]
    if len(tags) == 1:
        return tags[0]
    return ""


def cirq_circuit_to_stim_circuit(
    circuit: cirq.AbstractCircuit,
    *,
    qubit_to_index_dict: Optional[Dict[cirq.Qid, int]] = None,
    tag_func: Callable[[cirq.Operation], str] = _forward_single_str_tag,
) -> stim.Circuit:
    """Converts a cirq circuit into an equivalent stim circuit.

    Not all circuits can be converted. In order for a circuit to be convertible, all of its operations must be
    convertible.

    An operation is convertible if:
        - It is a stabilizer gate or probabilistic Pauli gate from cirq
            - cirq.H
            - cirq.S
            - cirq.X
            - cirq.X**0.5
            - cirq.CNOT
            - cirq.ResetChannel()
            - cirq.X.with_probability(p)
            - cirq.DepolarizingChannel(p, n_qubits=1 or 2)
            - etc
        - Or it has a _decompose_ method that yields convertible operations.
        - Or it has a correctly implemented _stim_conversion_ method.

    Args:
        circuit: The circuit to convert.
        qubit_to_index_dict: Optional. Which integer each qubit should get mapped to. If not specified, defaults to
            indexing qubits in the circuit in sorted order.
        tag_func: Controls the tag attached to the stim instructions the cirq operation turns
            into. If not specified, defaults to checking for string tags on the circuit operation
            and if there is exactly one string tag then using that tag (otherwise not specifying a
            tag).

    Returns:
        The converted circuit.

    Examples:
        >>> import cirq, stimcirq
        >>> a = cirq.NamedQubit("zero")
        >>> b = cirq.NamedQubit("two")
        >>> stimcirq.cirq_circuit_to_stim_circuit(cirq.Circuit(
        ...     cirq.Moment(cirq.H(a)),
        ...     cirq.Moment(cirq.CNOT(a, b)),
        ...     cirq.Moment(
        ...         cirq.X(a).with_probability(0.25),
        ...         cirq.Z(b).with_probability(0.25),
        ...     ),
        ...     cirq.Moment(),
        ...     cirq.Moment(),
        ...     cirq.Moment(cirq.DepolarizingChannel(0.125, n_qubits=2).on(b, a)),
        ...     cirq.Moment(cirq.measure(a, b)),
        ... ), qubit_to_index_dict={a: 0, b: 2})
        stim.Circuit('''
            H 0
            TICK
            CX 0 2
            TICK
            X_ERROR(0.25) 0
            Z_ERROR(0.25) 2
            TICK
            TICK
            TICK
            DEPOLARIZE2(0.125) 2 0
            TICK
            M 0 2
            TICK
        ''')

    Here is an example of a _stim_conversion_ method:

        def _stim_conversion_(
                self,

                # The stim circuit being built. Add onto it.
                edit_circuit: stim.Circuit,

                # Metadata about measurement groupings needed by stimcirq.StimSampler.
                # If your gate contains a measurement, it has to append how many qubits
                # that measurement measures (and its key) into this list.
                edit_measurement_key_lengths: List[Tuple[str, int]],

                # The indices of qubits the gate is operating on.
                targets: List[int],

                # A custom string associated with the operation, which can be tagged
                # onto any operations appended to the stim circuit.
                tag: str,

                # Forward compatibility with future arguments.
                **kwargs):

            edit_circuit.append_operation("H", targets)
    """
    return cirq_circuit_to_stim_data(circuit, q2i=qubit_to_index_dict, flatten=False, tag_func=tag_func)[0]


def cirq_circuit_to_stim_data(
    circuit: cirq.AbstractCircuit,
    *,
    q2i: Optional[Dict[cirq.Qid, int]] = None,
    flatten: bool = False,
    tag_func: Callable[[cirq.Operation], str] = _forward_single_str_tag,
) -> Tuple[stim.Circuit, List[Tuple[str, int]]]:
    """Converts a Cirq circuit into a Stim circuit and also metadata about where measurements go."""
    if q2i is None:
        q2i = {q: i for i, q in enumerate(sorted(circuit.all_qubits()))}
    helper = CirqToStimHelper(tag_func=tag_func)
    helper.q2i = q2i
    helper.flatten = flatten

    for q in sorted(circuit.all_qubits()):
        if isinstance(q, cirq.LineQubit):
            i = q2i[q]
            if i != q.x:
                helper.out.append_operation("QUBIT_COORDS", [i], [q.x])
        elif isinstance(q, cirq.GridQubit):
            helper.out.append_operation("QUBIT_COORDS", [q2i[q]], [q.row, q.col])

    helper.process_moments(circuit)
    return helper.out, helper.key_out


StimTypeHandler = Callable[[stim.Circuit, cirq.Gate, List[int], str], None]


@functools.lru_cache(maxsize=1)
def gate_to_stim_append_func() -> Dict[cirq.Gate, Callable[[stim.Circuit, List[int], str], None]]:
    """A dictionary mapping specific gate instances to stim circuit appending functions."""
    x = (cirq.X, False)
    y = (cirq.Y, False)
    z = (cirq.Z, False)
    nx = (cirq.X, True)
    ny = (cirq.Y, True)
    nz = (cirq.Z, True)

    def do_nothing(_gates, _targets, tag):
        pass

    def use(
        *gates: str, individuals: Sequence[Tuple[str, int]] = ()
    ) -> Callable[[stim.Circuit, List[int], str], None]:
        if len(gates) == 1 and not individuals:
            (g,) = gates
            return lambda c, t, tag: c.append(g, t, tag=tag)

        if not individuals:

            def do(c, t, tag: str):
                for g in gates:
                    c.append(g, t, tag=tag)

        else:

            def do(c, t, tag: str):
                for g in gates:
                    c.append(g, t, tag=tag)
                for g, k in individuals:
                    c.append(g, [t[k]], tag=tag)

        return do

    sqcg = cirq.SingleQubitCliffordGate.from_xz_map
    paulis = cast(List[cirq.Pauli], [cirq.X, cirq.Y, cirq.Z])

    return {
        cirq.ResetChannel(): use("R"),
        # Identities.
        cirq.I: use("I"),
        cirq.H ** 0: do_nothing,
        cirq.X ** 0: do_nothing,
        cirq.Y ** 0: do_nothing,
        cirq.Z ** 0: do_nothing,
        cirq.ISWAP ** 0: do_nothing,
        cirq.SWAP ** 0: do_nothing,
        # Common named gates.
        cirq.H: use("H"),
        cirq.X: use("X"),
        cirq.Y: use("Y"),
        cirq.Z: use("Z"),
        cirq.X ** 0.5: use("SQRT_X"),
        cirq.X ** -0.5: use("SQRT_X_DAG"),
        cirq.Y ** 0.5: use("SQRT_Y"),
        cirq.Y ** -0.5: use("SQRT_Y_DAG"),
        cirq.Z ** 0.5: use("SQRT_Z"),
        cirq.Z ** -0.5: use("SQRT_Z_DAG"),
        cirq.CNOT: use("CNOT"),
        cirq.CZ: use("CZ"),
        cirq.ISWAP: use("ISWAP"),
        cirq.ISWAP ** -1: use("ISWAP_DAG"),
        cirq.ISWAP ** 2: use("Z"),
        cirq.SWAP: use("SWAP"),
        cirq.X.controlled(1): use("CX"),
        cirq.Y.controlled(1): use("CY"),
        cirq.Z.controlled(1): use("CZ"),
        cirq.XX ** 0.5: use("SQRT_XX"),
        cirq.YY ** 0.5: use("SQRT_YY"),
        cirq.ZZ ** 0.5: use("SQRT_ZZ"),
        cirq.XX ** -0.5: use("SQRT_XX_DAG"),
        cirq.YY ** -0.5: use("SQRT_YY_DAG"),
        cirq.ZZ ** -0.5: use("SQRT_ZZ_DAG"),
        # All 24 cirq.SingleQubitCliffordGate instances.
        sqcg(x, y): use("SQRT_X_DAG"),
        sqcg(x, ny): use("SQRT_X"),
        sqcg(nx, y): use("H_YZ"),
        sqcg(nx, ny): use("H_NYZ"),
        sqcg(x, z): do_nothing,
        sqcg(x, nz): use("X"),
        sqcg(nx, z): use("Z"),
        sqcg(nx, nz): use("Y"),
        sqcg(y, x): use("C_XYZ"),
        sqcg(y, nx): use("C_XYNZ"),
        sqcg(ny, x): use("C_XNYZ"),
        sqcg(ny, nx): use("C_NXYZ"),
        sqcg(y, z): use("S"),
        sqcg(y, nz): use("H_XY"),
        sqcg(ny, z): use("S_DAG"),
        sqcg(ny, nz): use("H_NXY"),
        sqcg(z, x): use("H"),
        sqcg(z, nx): use("SQRT_Y_DAG"),
        sqcg(nz, x): use("SQRT_Y"),
        sqcg(nz, nx): use("H_NXZ"),
        sqcg(z, y): use("C_ZYX"),
        sqcg(z, ny): use("C_ZNYX"),
        sqcg(nz, y): use("C_ZYNX"),
        sqcg(nz, ny): use("C_NZYX"),
        # All 36 cirq.PauliInteractionGate instances.
        **{
            cirq.PauliInteractionGate(p0, s0, p1, s1): use(
                f"{p0}C{p1}", individuals=[(str(p1), 1)] * s0 + [(str(p0), 0)] * s1
            )
            for p0, s0, p1, s1 in itertools.product(paulis, [False, True], repeat=2)
        },
    }


@functools.lru_cache()
def gate_type_to_stim_append_func() -> Dict[Type[cirq.Gate], StimTypeHandler]:
    """A dictionary mapping specific gate types to stim circuit appending functions."""
    return {
        cirq.ControlledGate: cast(StimTypeHandler, _stim_append_controlled_gate),
        cirq.DensePauliString: cast(StimTypeHandler, _stim_append_dense_pauli_string_gate),
        cirq.MutableDensePauliString: cast(StimTypeHandler, _stim_append_dense_pauli_string_gate),
        cirq.AsymmetricDepolarizingChannel: cast(
            StimTypeHandler, _stim_append_asymmetric_depolarizing_channel
        ),
        cirq.BitFlipChannel: lambda c, g, t, tag: c.append(
            "X_ERROR", t, cast(cirq.BitFlipChannel, g).p, tag=tag
        ),
        cirq.PhaseFlipChannel: lambda c, g, t, tag: c.append(
            "Z_ERROR", t, cast(cirq.PhaseFlipChannel, g).p, tag=tag
        ),
        cirq.PhaseDampingChannel: lambda c, g, t, tag: c.append(
            "Z_ERROR", t, 0.5 - math.sqrt(1 - cast(cirq.PhaseDampingChannel, g).gamma) / 2, tag=tag
        ),
        cirq.RandomGateChannel: cast(StimTypeHandler, _stim_append_random_gate_channel),
        cirq.DepolarizingChannel: cast(StimTypeHandler, _stim_append_depolarizing_channel),
    }


def _stim_append_measurement_gate(
    circuit: stim.Circuit, gate: cirq.MeasurementGate, targets: List[int], tag: str
):
    for i, b in enumerate(gate.invert_mask):
        if b:
            targets[i] = stim.target_inv(targets[i])
    circuit.append("M", targets, tag=tag)


def _stim_append_pauli_measurement_gate(
    circuit: stim.Circuit, gate: cirq.PauliMeasurementGate, targets: List[int], tag: str
):
    obs: cirq.DensePauliString = gate.observable()

    # Convert to stim Pauli product targets.
    if len(targets) == 0:
        raise NotImplementedError(f"len(targets)={len(targets)} == 0")
    new_targets = []
    for t, p in zip(targets, obs.pauli_mask):
        if p == 1:
            t = stim.target_x(t, invert=not new_targets and obs.coefficient == -1)
        elif p == 2:
            t = stim.target_y(t, invert=not new_targets and obs.coefficient == -1)
        elif p == 3:
            t = stim.target_z(t, invert=not new_targets and obs.coefficient == -1)
        else:
            raise NotImplementedError(f"obs={obs!r}")
        new_targets.append(t)
        new_targets.append(stim.target_combiner())
    new_targets.pop()

    # Inverted result?
    if obs.coefficient != 1 and obs.coefficient != -1:
        raise NotImplementedError(f"obs.coefficient={obs.coefficient!r} not in [1, -1]")

    circuit.append("MPP", new_targets, tag=tag)


def _stim_append_spp_gate(
    circuit: stim.Circuit, gate: cirq.PauliStringPhasorGate, targets: List[int], tag: str
):
    obs: cirq.DensePauliString = gate.dense_pauli_string
    a = gate.exponent_neg
    b = gate.exponent_pos
    d = (a - b) % 2
    if obs.coefficient == -1:
        d += 1
        d %= 2
    if d != 0.5 and d != 1.5:
        return False

    new_targets = []
    for t, p in zip(targets, obs.pauli_mask):
        if p:
            new_targets.append(stim.target_pauli(t, p))
            new_targets.append(stim.target_combiner())
    if len(new_targets) == 0:
        return False
    new_targets.pop()

    circuit.append("SPP" if d == 0.5 else "SPP_DAG", new_targets, tag=tag)
    return True


def _stim_append_dense_pauli_string_gate(
    c: stim.Circuit, g: cirq.BaseDensePauliString, t: List[int], tag: str
):
    gates = [None, "X", "Y", "Z"]
    for p, k in zip(g.pauli_mask, t):
        if p:
            c.append(gates[p], [k], tag=tag)


def _stim_append_asymmetric_depolarizing_channel(
    c: stim.Circuit, g: cirq.AsymmetricDepolarizingChannel, t: List[int], tag: str
):
    if cirq.num_qubits(g) == 1:
        c.append("PAULI_CHANNEL_1", t, [g.p_x, g.p_y, g.p_z], tag=tag)
    elif cirq.num_qubits(g) == 2:
        c.append(
            "PAULI_CHANNEL_2",
            t,
            [
                g.error_probabilities.get('IX', 0),
                g.error_probabilities.get('IY', 0),
                g.error_probabilities.get('IZ', 0),
                g.error_probabilities.get('XI', 0),
                g.error_probabilities.get('XX', 0),
                g.error_probabilities.get('XY', 0),
                g.error_probabilities.get('XZ', 0),
                g.error_probabilities.get('YI', 0),
                g.error_probabilities.get('YX', 0),
                g.error_probabilities.get('YY', 0),
                g.error_probabilities.get('YZ', 0),
                g.error_probabilities.get('ZI', 0),
                g.error_probabilities.get('ZX', 0),
                g.error_probabilities.get('ZY', 0),
                g.error_probabilities.get('ZZ', 0),
            ],
            tag=tag,
        )
    else:
        raise NotImplementedError(f'cirq-to-stim gate {g!r}')


def _stim_append_depolarizing_channel(c: stim.Circuit, g: cirq.DepolarizingChannel, t: List[int], tag: str):
    if g.num_qubits() == 1:
        c.append("DEPOLARIZE1", t, g.p, tag=tag)
    elif g.num_qubits() == 2:
        c.append("DEPOLARIZE2", t, g.p, tag=tag)
    else:
        raise TypeError(f"Don't know how to turn {g!r} into Stim operations.")


def _stim_append_controlled_gate(c: stim.Circuit, g: cirq.ControlledGate, t: List[int], tag: str):
    if isinstance(g.sub_gate, cirq.BaseDensePauliString) and g.num_controls() == 1:
        gates = [None, "CX", "CY", "CZ"]
        for p, k in zip(g.sub_gate.pauli_mask, t[1:]):
            if p:
                c.append(gates[p], [t[0], k], tag=tag)
        if g.sub_gate.coefficient == 1j:
            c.append("S", t[:1], tag=tag)
        elif g.sub_gate.coefficient == -1:
            c.append("Z", t[:1], tag=tag)
        elif g.sub_gate.coefficient == -1j:
            c.append("S_DAG", t[:1], tag=tag)
        elif g.sub_gate.coefficient == 1:
            pass
        else:
            raise TypeError(f"Phase kickback from {g!r} isn't a stabilizer operation.")
        return

    raise TypeError(f"Don't know how to turn controlled gate {g!r} into Stim operations.")


def _stim_append_random_gate_channel(c: stim.Circuit, g: cirq.RandomGateChannel, t: List[int], tag: str):
    if g.sub_gate in [cirq.X, cirq.Y, cirq.Z, cirq.I]:
        c.append(f"{g.sub_gate}_ERROR", t, g.probability, tag=tag)
    elif isinstance(g.sub_gate, IIGate):
        c.append(f"II_ERROR", t, g.probability, tag=tag)
    elif isinstance(g.sub_gate, cirq.DensePauliString):
        target_p = [None, stim.target_x, stim.target_y, stim.target_z]
        pauli_targets = [target_p[p](t) for t, p in zip(t, g.sub_gate.pauli_mask) if p]
        c.append(f"CORRELATED_ERROR", pauli_targets, g.probability, tag=tag)
    else:
        raise NotImplementedError(
            f"Don't know how to turn probabilistic {g!r} into Stim operations."
        )


class CirqToStimHelper:
    def __init__(self, tag_func: Callable[[cirq.Operation], str]):
        self.key_out: List[Tuple[str, int]] = []
        self.out = stim.Circuit()
        self.q2i = {}
        self.have_seen_loop = False
        self.flatten = False
        self.tag_func = tag_func

    def process_circuit_operation_into_repeat_block(self, op: cirq.CircuitOperation, tag: str) -> None:
        if self.flatten or op.repetitions == 1:
            moments = cirq.unroll_circuit_op(cirq.Circuit(op), deep=False, tags_to_check=None).moments
            self.process_moments(moments)
            self.out = self.out[:-1] # Remove a trailing TICK (to avoid double TICK)
            return

        child = CirqToStimHelper(tag_func=self.tag_func)
        child.key_out = self.key_out
        child.q2i = self.q2i
        child.have_seen_loop = True
        self.have_seen_loop = True
        child.process_moments(op.transform_qubits(lambda q: op.qubit_map.get(q, q)).circuit)
        self.out.append(stim.CircuitRepeatBlock(op.repetitions, child.out, tag=tag))

    def process_operations(self, operations: Iterable[cirq.Operation]) -> None:
        g2f = gate_to_stim_append_func()
        t2f = gate_type_to_stim_append_func()
        for op in operations:
            assert isinstance(op, cirq.Operation)
            tag = self.tag_func(op)
            op = op.untagged
            gate = op.gate
            targets = [self.q2i[q] for q in op.qubits]

            custom_method = getattr(
                op, '_stim_conversion_', getattr(gate, '_stim_conversion_', None)
            )
            if custom_method is not None:
                custom_method(
                    dont_forget_your_star_star_kwargs=True,
                    edit_circuit=self.out,
                    edit_measurement_key_lengths=self.key_out,
                    targets=targets,
                    have_seen_loop=self.have_seen_loop,
                    tag=tag,
                )
                continue

            if isinstance(op, cirq.CircuitOperation):
                self.process_circuit_operation_into_repeat_block(op, tag=tag)
                continue

            # Special case measurement, because of its metadata.
            if isinstance(gate, cirq.PauliStringPhasorGate):
                if _stim_append_spp_gate(self.out, gate, targets, tag=tag):
                    continue
            if isinstance(gate, cirq.PauliMeasurementGate):
                self.key_out.append((gate.key, len(targets)))
                _stim_append_pauli_measurement_gate(self.out, gate, targets, tag=tag)
                continue
            if isinstance(gate, cirq.MeasurementGate):
                self.key_out.append((gate.key, len(targets)))
                _stim_append_measurement_gate(self.out, gate, targets, tag=tag)
                continue

            # Look for recognized gate values like cirq.H.
            val_append_func = g2f.get(gate)
            if val_append_func is not None:
                val_append_func(self.out, targets, tag=tag)
                continue

            # Look for recognized gate types like cirq.DepolarizingChannel.
            type_append_func = t2f.get(type(gate))
            if type_append_func is not None:
                type_append_func(self.out, gate, targets, tag=tag)
                continue

            # Ask unrecognized operations to decompose themselves into simpler operations.
            try:
                self.process_operations(cirq.decompose_once(op))
            except TypeError as ex:
                raise TypeError(
                    f"Don't know how to translate {op!r} into stim gates.\n"
                    f"- It doesn't have a _decompose_ method that returns stim-compatible operations.\n"
                    f"- It doesn't have a _stim_conversion_ method.\n"
                ) from ex

    def process_moment(self, moment: cirq.Moment):
        length_before = len(self.out)
        self.process_operations(moment)

        # Append a TICK, unless it was already handled by an internal REPEAT block.
        if length_before == len(self.out) or not isinstance(self.out[-1], stim.CircuitRepeatBlock):
            self.out.append("TICK", [])

    def process_moments(self, moments: Iterable[cirq.Moment]):
        for moment in moments:
            self.process_moment(moment)



import itertools
from typing import cast

import cirq
import pytest
import stim

import stimcirq
from ._stim_to_cirq import CircuitTranslationTracker


def test_stim_circuit_to_cirq_circuit():
    circuit = stimcirq.stim_circuit_to_cirq_circuit(
        stim.Circuit(
            """
        X 0
        CNOT 0 1
        X_ERROR(0.125) 0 1
        CORRELATED_ERROR(0.25) X0 Y1 Z2
        M 1
        M !1
        TICK
        MR 0 !1
    """
        )
    )
    a, b, c = cirq.LineQubit.range(3)
    assert circuit == cirq.Circuit(
        cirq.Moment(cirq.X(a)),
        cirq.Moment(cirq.CNOT(a, b)),
        cirq.Moment(cirq.X.with_probability(0.125).on(a), cirq.X.with_probability(0.125).on(b)),
        cirq.Moment(cirq.PauliString({a: cirq.X, b: cirq.Y, c: cirq.Z}).with_probability(0.25)),
        cirq.Moment(cirq.measure(b, key="0")),
        cirq.Moment(cirq.measure(b, key="1", invert_mask=(True,))),
        cirq.Moment(
            stimcirq.MeasureAndOrResetGate(
                measure=True, reset=True, basis='Z', invert_measure=False, key='2'
            ).on(a),
            stimcirq.MeasureAndOrResetGate(
                measure=True, reset=True, basis='Z', invert_measure=True, key='3'
            ).on(b),
        ),
    )


def assert_circuits_are_equivalent_and_convert(
    cirq_circuit: cirq.Circuit, stim_circuit: stim.Circuit
):
    assert cirq_circuit == stimcirq.stim_circuit_to_cirq_circuit(stim_circuit)
    assert stim_circuit == stimcirq.cirq_circuit_to_stim_circuit(cirq_circuit)


@pytest.mark.parametrize(
    "name",
    [
        k
        for k, v in CircuitTranslationTracker.get_handler_table().items()
        if isinstance(v, CircuitTranslationTracker.OneToOneGateHandler)
    ],
)
def test_gates_converted_using_OneToOneGateHandler(name: str):
    handler = cast(
        CircuitTranslationTracker.OneToOneGateHandler,
        CircuitTranslationTracker.get_handler_table()[name],
    )
    gate = handler.gate
    n = cirq.num_qubits(gate)
    qs = cirq.LineQubit.range(n)

    cirq_original = cirq.Circuit(gate.on(*qs))
    stim_targets = " ".join(str(e) for e in range(n))
    stim_original = stim.Circuit(f"{name} {stim_targets}\nTICK")
    assert_circuits_are_equivalent_and_convert(cirq_original, stim_original)


@pytest.mark.parametrize(
    "name",
    [
        k
        for k, v in CircuitTranslationTracker.get_handler_table().items()
        if isinstance(v, CircuitTranslationTracker.SweepableGateHandler)
    ],
)
def test_gates_converted_using_SweepableGateHandler(name: str):
    handler = cast(
        CircuitTranslationTracker.SweepableGateHandler,
        CircuitTranslationTracker.get_handler_table()[name],
    )
    gate = handler.gate
    n = cirq.num_qubits(gate)
    qs = cirq.LineQubit.range(n)

    # Without sweeping.
    cirq_original = cirq.Circuit(gate.on(*qs))
    stim_targets = " ".join(str(e) for e in range(n))
    stim_original = stim.Circuit(f"{name} {stim_targets}\nTICK")
    assert_circuits_are_equivalent_and_convert(cirq_original, stim_original)

    # With sweeping.
    q = qs[0]
    cirq_original = cirq.Circuit(
        stimcirq.SweepPauli(
            stim_sweep_bit_index=3, cirq_sweep_symbol="sweep[3]", pauli=handler.pauli_gate
        ).on(q)
    )
    if name.startswith("C") or name.startswith("Z"):
        stim_original = stim.Circuit(f"{name} sweep[3] 0\nTICK")
        assert_circuits_are_equivalent_and_convert(cirq_original, stim_original)
    else:
        stim_original = stim.Circuit(f"{name} 0 sweep[3]\nTICK")
        # Round trip loses distinction between ZCX and XCZ.
        assert stimcirq.stim_circuit_to_cirq_circuit(stim_original) == cirq_original


@pytest.mark.parametrize(
    "name,probability",
    itertools.product(
        [
            k
            for k, v in CircuitTranslationTracker.get_handler_table().items()
            if isinstance(v, CircuitTranslationTracker.OneToOneNoisyGateHandler)
        ],
        [0, 0.125],
    ),
)
def test_gates_converted_using_OneToOneNoisyGateHandler(name: str, probability: float):
    handler = cast(
        CircuitTranslationTracker.OneToOneNoisyGateHandler,
        CircuitTranslationTracker.get_handler_table()[name],
    )
    gate = handler.prob_to_gate(probability)
    n = cirq.num_qubits(gate)
    qs = cirq.LineQubit.range(n)

    cirq_original = cirq.Circuit(gate.on(*qs))
    stim_targets = " ".join(str(e) for e in range(n))
    stim_original = stim.Circuit(f"{name}({probability}) {stim_targets}\nTICK")
    assert_circuits_are_equivalent_and_convert(cirq_original, stim_original)


@pytest.mark.parametrize(
    "name,invert,probability",
    itertools.product(
        [
            k
            for k, v in CircuitTranslationTracker.get_handler_table().items()
            if isinstance(v, CircuitTranslationTracker.OneToOneMeasurementHandler)
        ],
        [False, True],
        [0, 0.125],
    ),
)
def test_gates_converted_using_OneToOneMeasurementHandler(
    name: str, invert: bool, probability: float
):
    handler = cast(
        CircuitTranslationTracker.OneToOneMeasurementHandler,
        CircuitTranslationTracker.get_handler_table()[name],
    )
    if handler.basis == 'Z' and not handler.reset and not probability:
        cirq_original = cirq.Circuit(
            cirq.measure(cirq.LineQubit(5), key="0", invert_mask=(1,) * invert)
        )
    else:
        cirq_original = cirq.Circuit(
            stimcirq.MeasureAndOrResetGate(
                basis=handler.basis,
                key="0",
                invert_measure=invert,
                reset=handler.reset,
                measure_flip_probability=probability,
                measure=handler.measure,
            ).on(cirq.LineQubit(5))
        )
    inverted = "!" if invert else ""
    p_str = f"({probability})" if probability else ""
    stim_original = stim.Circuit(f"QUBIT_COORDS(5) 0\n{name}{p_str} {inverted}0\nTICK")
    assert_circuits_are_equivalent_and_convert(cirq_original, stim_original)


def test_round_trip_preserves_moment_structure():
    a, b = cirq.LineQubit.range(2)
    circuit = cirq.Circuit(
        cirq.Moment(),
        cirq.Moment(cirq.H(a)),
        cirq.Moment(cirq.H(b)),
        cirq.Moment(),
        cirq.Moment(),
        cirq.Moment(cirq.CNOT(a, b)),
        cirq.Moment(),
    )
    converted = stimcirq.cirq_circuit_to_stim_circuit(circuit)
    restored = stimcirq.stim_circuit_to_cirq_circuit(converted)
    assert restored == circuit


def test_circuit_diagram():
    cirq.testing.assert_has_diagram(
        stimcirq.stim_circuit_to_cirq_circuit(
            stim.Circuit(
                """
            M 9
            MX 9
            MY 9
            MZ 9
            R 9
            RX 9
            RY 9
            RZ 9
            MR 9
            MRX 9
            MRY 9
            MRZ 9
        """
            )
        ),
        """
9: ---M('0')---MX('1')---MY('2')---M('3')---R---RX---RY---R---MR('4')---MRX('5')---MRY('6')---MR('7')---
        """,
        use_unicode_characters=False,
    )


def test_all_known_gates_explicitly_handled():
    gates = stim.gate_data()
    handled = CircuitTranslationTracker.get_handler_table().keys()
    for gate_name in gates:
        assert gate_name in handled, gate_name


def test_line_grid_qubit_round_trip():
    c = cirq.Circuit(
        cirq.H(cirq.LineQubit(101)),
        cirq.Z(cirq.LineQubit(200.5)),
        cirq.X(cirq.GridQubit(2, 3.5)),
        cirq.Y(cirq.GridQubit(-2, 5)),
    )
    s = stimcirq.cirq_circuit_to_stim_circuit(c)
    assert s == stim.Circuit(
        """
        QUBIT_COORDS(-2, 5) 0
        QUBIT_COORDS(2, 3.5) 1
        QUBIT_COORDS(101) 2
        QUBIT_COORDS(200.5) 3
        H 2
        Z 3
        X 1
        Y 0
        TICK
    """
    )
    c2 = stimcirq.stim_circuit_to_cirq_circuit(s)
    for q in c2.all_qubits():
        if q == cirq.LineQubit(101):
            assert isinstance(cast(cirq.LineQubit, q).x, int)
        if q == cirq.GridQubit(2, 3.5):
            assert isinstance(cast(cirq.GridQubit, q).row, int)
    assert c2 == c

    assert (
        stimcirq.stim_circuit_to_cirq_circuit(
            stim.Circuit(
                """
        QUBIT_COORDS(10) 0
        SHIFT_COORDS(1, 2, 3)
        QUBIT_COORDS(20, 30) 1
        H 0 1
    """
            )
        )
        == cirq.Circuit(
            stimcirq.ShiftCoordsAnnotation([1, 2, 3]),
            cirq.H(cirq.LineQubit(10)),
            cirq.H(cirq.GridQubit(21, 32)),
        )
    )


def test_noisy_measurements():
    s = stim.Circuit(
        """
        MX(0.125) 0
        MY(0.125) 1
        MZ(0.125) 2
        MRX(0.125) 3
        MRY(0.125) 4
        MRZ(0.25) 5
        TICK
    """
    )
    c = stimcirq.stim_circuit_to_cirq_circuit(s)
    assert c == cirq.Circuit(
        stimcirq.MeasureAndOrResetGate(
            measure=True,
            reset=False,
            basis='X',
            invert_measure=False,
            key='0',
            measure_flip_probability=0.125,
        ).on(cirq.LineQubit(0)),
        stimcirq.MeasureAndOrResetGate(
            measure=True,
            reset=False,
            basis='Y',
            invert_measure=False,
            key='1',
            measure_flip_probability=0.125,
        ).on(cirq.LineQubit(1)),
        stimcirq.MeasureAndOrResetGate(
            measure=True,
            reset=False,
            basis='Z',
            invert_measure=False,
            key='2',
            measure_flip_probability=0.125,
        ).on(cirq.LineQubit(2)),
        stimcirq.MeasureAndOrResetGate(
            measure=True,
            reset=True,
            basis='X',
            invert_measure=False,
            key='3',
            measure_flip_probability=0.125,
        ).on(cirq.LineQubit(3)),
        stimcirq.MeasureAndOrResetGate(
            measure=True,
            reset=True,
            basis='Y',
            invert_measure=False,
            key='4',
            measure_flip_probability=0.125,
        ).on(cirq.LineQubit(4)),
        stimcirq.MeasureAndOrResetGate(
            measure=True,
            reset=True,
            basis='Z',
            invert_measure=False,
            key='5',
            measure_flip_probability=0.25,
        ).on(cirq.LineQubit(5)),
    )
    assert stimcirq.cirq_circuit_to_stim_circuit(c) == s
    cirq.testing.assert_equivalent_repr(c, global_vals={'stimcirq': stimcirq})


def test_convert_mpp():
    s = stim.Circuit(
        """
        MPP X2*Y5*Z3
        TICK
        MPP Y2 X1 Z3*Z4
        X 0
        TICK
    """
    )
    c = cirq.Circuit(
        cirq.Moment(
            cirq.PauliMeasurementGate(cirq.DensePauliString("XYZ", coefficient=+1), key='0').on(
                cirq.LineQubit(2), cirq.LineQubit(5), cirq.LineQubit(3)
            )
        ),
        cirq.Moment(
            cirq.PauliMeasurementGate(cirq.DensePauliString("Y"), key='1').on(cirq.LineQubit(2)),
            cirq.PauliMeasurementGate(cirq.DensePauliString("X"), key='2').on(cirq.LineQubit(1)),
            cirq.PauliMeasurementGate(cirq.DensePauliString("ZZ"), key='3').on(
                cirq.LineQubit(3), cirq.LineQubit(4)
            ),
            cirq.X(cirq.LineQubit(0)),
        ),
    )
    assert_circuits_are_equivalent_and_convert(c, s)
    cirq.testing.assert_has_diagram(
        c,
        """
0: ---------------X-----------

1: ---------------M(X)('2')---

2: ---M(X)('0')---M(Y)('1')---
      |
3: ---M(Z)--------M(Z)('3')---
      |           |
4: ---|-----------M(Z)--------
      |
5: ---M(Y)--------------------
        """,
        use_unicode_characters=False,
    )


def test_convert_detector():
    s = stim.Circuit(
        """
        H 0
        TICK
        CNOT 0 1
        TICK
        M 1 0
        DETECTOR(2, 3, 5) rec[-1] rec[-2]
        TICK
    """
    )
    a, b = cirq.LineQubit.range(2)
    c = cirq.Circuit(
        cirq.H(a),
        cirq.CNOT(a, b),
        cirq.Moment(
            cirq.measure(b, key='0'),
            cirq.measure(a, key='1'),
            stimcirq.DetAnnotation(parity_keys=['0', '1'], coordinate_metadata=(2, 3, 5)),
        ),
    )
    cirq.testing.assert_has_diagram(
        c,
        """
0: ---H---@---M('1')---------
          |
1: -------X---M('0')---------
              Det('0','1')
        """,
        use_unicode_characters=False,
    )
    assert_circuits_are_equivalent_and_convert(c, s)


def test_convert_observable():
    s = stim.Circuit(
        """
        H 0
        TICK
        CNOT 0 1
        TICK
        M 1 0
        OBSERVABLE_INCLUDE(5) rec[-1] rec[-2]
        TICK
    """
    )
    a, b = cirq.LineQubit.range(2)
    c = cirq.Circuit(
        cirq.H(a),
        cirq.CNOT(a, b),
        cirq.Moment(
            cirq.measure(b, key='0'),
            cirq.measure(a, key='1'),
            stimcirq.CumulativeObservableAnnotation(parity_keys=['0', '1'], observable_index=5),
        ),
    )
    cirq.testing.assert_has_diagram(
        c,
        """
0: ---H---@---M('1')----------
          |
1: -------X---M('0')----------
              Obs5('0','1')
        """,
        use_unicode_characters=False,
    )
    assert_circuits_are_equivalent_and_convert(c, s)


def test_sweep_target():
    stim_circuit = stim.Circuit(
        """
        CX sweep[2] 0
        CY sweep[3] 1
        CZ sweep[5] 2 sweep[7] 3
        TICK
    """
    )
    a, b, c, d = cirq.LineQubit.range(4)
    cirq_circuit = cirq.Circuit(
        stimcirq.SweepPauli(stim_sweep_bit_index=2, cirq_sweep_symbol="sweep[2]", pauli=cirq.X).on(
            a
        ),
        stimcirq.SweepPauli(stim_sweep_bit_index=3, cirq_sweep_symbol="sweep[3]", pauli=cirq.Y).on(
            b
        ),
        stimcirq.SweepPauli(stim_sweep_bit_index=5, cirq_sweep_symbol="sweep[5]", pauli=cirq.Z).on(
            c
        ),
        stimcirq.SweepPauli(stim_sweep_bit_index=7, cirq_sweep_symbol="sweep[7]", pauli=cirq.Z).on(
            d
        ),
    )
    cirq.testing.assert_has_diagram(
        cirq_circuit,
        """
0: ---X^sweep[2]='sweep[2]'---

1: ---Y^sweep[3]='sweep[3]'---

2: ---Z^sweep[5]='sweep[5]'---

3: ---Z^sweep[7]='sweep[7]'---
        """,
        use_unicode_characters=False,
    )
    assert_circuits_are_equivalent_and_convert(cirq_circuit, stim_circuit)


def test_convert_repeat_simple():
    stim_circuit = stim.Circuit(
        """
        REPEAT 1000000 {
            H 0
            TICK
            CNOT 0 1
            TICK
        }
        M 0
        TICK
    """
    )
    a, b = cirq.LineQubit.range(2)
    cirq_circuit = cirq.Circuit(
        cirq.CircuitOperation(
            cirq.FrozenCircuit(cirq.H(a), cirq.CNOT(a, b)),
            repetitions=1000000,
            use_repetition_ids=False,
        ),
        cirq.measure(a, key="0"),
    )
    assert stimcirq.stim_circuit_to_cirq_circuit(stim_circuit) == cirq_circuit
    assert stimcirq.cirq_circuit_to_stim_circuit(cirq_circuit) == stim_circuit


def test_convert_repeat_measurements():
    stim_circuit = stim.Circuit(
        """
        QUBIT_COORDS(2, 3) 0
        QUBIT_COORDS(4, 5) 1
        REPEAT 1000000 {
            H 0
            TICK
            CNOT 0 1
            TICK
            M 0 1
            DETECTOR(5) rec[-1] rec[-2]
            SHIFT_COORDS(2)
            TICK
        }
        M 0
        DETECTOR(7) rec[-1] rec[-3]
        TICK
    """
    )
    a, b = cirq.GridQubit(2, 3), cirq.GridQubit(4, 5)
    cirq_circuit = cirq.Circuit(
        cirq.Moment(
            cirq.CircuitOperation(
                cirq.FrozenCircuit(
                    cirq.Moment(cirq.H(a)),
                    cirq.Moment(cirq.CNOT(a, b)),
                    cirq.Moment(
                        cirq.measure(a, key=cirq.MeasurementKey("0")),
                        cirq.measure(b, key=cirq.MeasurementKey("1")),
                        stimcirq.DetAnnotation(relative_keys=[-1, -2], coordinate_metadata=[5]),
                        stimcirq.ShiftCoordsAnnotation([2]),
                    ),
                ),
                repetitions=1000000,
                use_repetition_ids=False,
            )
        ),
        cirq.Moment(
            cirq.measure(a, key="2000000"),
            stimcirq.DetAnnotation(relative_keys=[-1, -3], coordinate_metadata=[7]),
        ),
    )
    assert stimcirq.stim_circuit_to_cirq_circuit(stim_circuit) == cirq_circuit
    assert stimcirq.cirq_circuit_to_stim_circuit(cirq_circuit) == stim_circuit


def test_single_repeat_loops_always_flattened():
    assert stimcirq.stim_circuit_to_cirq_circuit(stim.Circuit(
        """
        REPEAT 1 {
            H 0
        }
        """)) == cirq.Circuit(cirq.H(cirq.LineQubit(0)))
    assert stimcirq.cirq_circuit_to_stim_circuit(cirq.Circuit(
        cirq.CircuitOperation(
            cirq.FrozenCircuit(
                cirq.H(cirq.LineQubit(0)),
            ),
            repetitions=1,
        ),
    )) == stim.Circuit("H 0\nTICK")


def test_circuit_operations_always_in_isolated_moment():
    stim_circuit = stim.Circuit(
        """
        REPEAT 2 {
            H 0 1
        }
        H 2 3
        """
    )
    cirq_circuit = stimcirq.stim_circuit_to_cirq_circuit(stim_circuit)
    qs = cirq.LineQubit.range(4)
    expected_cirq_circuit = cirq.Circuit(
        cirq.CircuitOperation(
            circuit=cirq.FrozenCircuit([cirq.H(qs[0]),cirq.H(qs[1])]),
            repetitions=2,
            use_repetition_ids=False
        ),
        cirq.Moment(cirq.H(qs[2]),cirq.H(qs[3])),
    )
    assert cirq_circuit == expected_cirq_circuit


@pytest.mark.skip(reason="blocked by https://github.com/quantumlib/Cirq/issues/6136")
def test_stim_circuit_to_cirq_circuit_mpad():
    stim_circuit = stim.Circuit("""
        MPAD 0 1
    """)
    cirq_circuit = stimcirq.stim_circuit_to_cirq_circuit(stim_circuit)
    assert cirq_circuit == cirq.Circuit(
        cirq.PauliMeasurementGate(cirq.DensePauliString(""), key="0").on(),
        cirq.PauliMeasurementGate(cirq.DensePauliString("", coefficient=-1), key="1").on(),
    )


def test_stim_circuit_to_cirq_circuit_mxx_myy_mzz():
    stim_circuit = stim.Circuit("""
        MXX 0 1
        MYY 0 !1
        MZZ !0 1
    """)
    cirq_circuit = stimcirq.stim_circuit_to_cirq_circuit(stim_circuit)
    a, b = cirq.LineQubit.range(2)
    assert cirq_circuit == cirq.Circuit(
        cirq.PauliMeasurementGate(cirq.DensePauliString("XX"), key='0').on(a, b),
        cirq.PauliMeasurementGate(cirq.DensePauliString("YY", coefficient=-1), key='1').on(a, b),
        cirq.PauliMeasurementGate(cirq.DensePauliString("ZZ", coefficient=-1), key='2').on(a, b),
    )
    assert stimcirq.cirq_circuit_to_stim_circuit(cirq_circuit) == stim.Circuit("""
        MPP X0*X1
        TICK
        MPP !Y0*Y1
        TICK
        MPP !Z0*Z1
        TICK
    """)


def test_stim_circuit_to_cirq_circuit_spp():
    stim_circuit = stim.Circuit("""
        SPP X1*Y2*Z3
        SPP !Y4
        SPP_DAG X5
        SPP_DAG !Z0
    """)
    cirq_circuit = stimcirq.stim_circuit_to_cirq_circuit(stim_circuit)
    assert cirq_circuit.to_text_diagram(use_unicode_characters=False).strip() == """
0: ---[Z]^0.5----

1: ---[X]--------
      |
2: ---[Y]--------
      |
3: ---[Z]^0.5----

4: ---[Y]^-0.5---

5: ---[X]^-0.5---
    """.strip()
    q0, q1, q2, q3, q4, q5 = cirq.LineQubit.range(6)
    assert cirq_circuit == cirq.Circuit([
        cirq.Moment(
            cirq.PauliStringPhasor(cirq.X(q1)*cirq.Y(q2)*cirq.Z(q3), exponent_neg=0.5),
            cirq.PauliStringPhasor(cirq.Y(q4), exponent_neg=0, exponent_pos=0.5),
            cirq.PauliStringPhasor(cirq.X(q5), exponent_neg=-0.5),
            cirq.PauliStringPhasor(cirq.Z(q0), exponent_neg=0, exponent_pos=-0.5),
        ),
    ])
    assert stimcirq.cirq_circuit_to_stim_circuit(cirq_circuit) == stim.Circuit("""
        SPP X1*Y2*Z3
        SPP_DAG Y4 X5
        SPP Z0
        TICK
    """)


def test_tags_convert():
    assert stimcirq.stim_circuit_to_cirq_circuit(stim.Circuit("""
        H[my_tag] 0
    """)) == cirq.Circuit(
        cirq.H(cirq.LineQubit(0)).with_tags('my_tag'),
    )


@pytest.mark.parametrize('gate', sorted(stim.gate_data().keys()))
def test_every_operation_converts_tags(gate: str):
    if gate in [
        "ELSE_CORRELATED_ERROR",
        "HERALDED_ERASE",
        "HERALDED_PAULI_CHANNEL_1",
        "TICK",
        "REPEAT",
        "MPAD",
        "QUBIT_COORDS",
    ]:
        pytest.skip()

    data = stim.gate_data(gate)
    stim_circuit = stim.Circuit()
    arg = None
    targets = [0, 1]
    if data.num_parens_arguments_range.start:
        arg = [2**-6] * data.num_parens_arguments_range.start
    if data.takes_pauli_targets:
        targets = [stim.target_x(0), stim.target_y(1)]
    if data.takes_measurement_record_targets and not data.is_unitary:
        stim_circuit.append("M", [0], tag='custom_tag')
        targets = [stim.target_rec(-1)]
    if gate == 'SHIFT_COORDS':
        targets = []
    if gate == 'OBSERVABLE_INCLUDE':
        arg = [1]
    stim_circuit.append(gate, targets, arg, tag='custom_tag')
    cirq_circuit = stimcirq.stim_circuit_to_cirq_circuit(stim_circuit)
    assert any(cirq_circuit.all_operations())
    for op in cirq_circuit.all_operations():
        assert op.tags == ('custom_tag',)
    restored_circuit = stimcirq.cirq_circuit_to_stim_circuit(cirq_circuit)
    assert restored_circuit.pop() == stim.CircuitInstruction("TICK")
    assert all(instruction.tag == 'custom_tag' for instruction in restored_circuit)
    if gate not in ['MXX', 'MYY', 'MZZ']:
        assert restored_circuit == stim_circuit


def test_loop_tagging():
    stim_circuit = stim.Circuit("""
        REPEAT[custom-tag] 5 {
            H[tag2] 0
            TICK
        }
    """)
    cirq_circuit = stimcirq.stim_circuit_to_cirq_circuit(stim_circuit)
    assert cirq_circuit == cirq.Circuit(
        cirq.CircuitOperation(
            cirq.FrozenCircuit(
                cirq.H(cirq.LineQubit(0)).with_tags('tag2'),
            ),
            repetitions=5,
            use_repetition_ids=False,
        ).with_tags('custom-tag')
    )
    restored_circuit = stimcirq.cirq_circuit_to_stim_circuit(cirq_circuit)
    assert restored_circuit == stim_circuit


def test_id_error_round_trip():
    stim_circuit = stim.Circuit("""
        I_ERROR 0
        I_ERROR(0.125, 0.25) 1
        II_ERROR(0.25, 0.125) 2 3
        II_ERROR 4 5
        TICK
    """)
    cirq_circuit = stimcirq.stim_circuit_to_cirq_circuit(stim_circuit)
    restored_circuit = stimcirq.cirq_circuit_to_stim_circuit(cirq_circuit)
    assert restored_circuit == stim_circuit

def test_round_trip_with_pauli_obs():
    stim_circuit = stim.Circuit("""
        QUBIT_COORDS(5, 5) 0
        R 0
        OBSERVABLE_INCLUDE(0) X0
        TICK
        H 0
        TICK
        M 0
        OBSERVABLE_INCLUDE(0) rec[-1]
        TICK
    """)
    cirq_circuit = stimcirq.stim_circuit_to_cirq_circuit(stim_circuit)
    restored_circuit = stimcirq.cirq_circuit_to_stim_circuit(cirq_circuit)
    assert restored_circuit == stim_circuit


__version__ = '1.16.dev1751065005'
from ._cirq_to_stim import cirq_circuit_to_stim_circuit
from ._cx_swap_gate import CXSwapGate
from ._cz_swap_gate import CZSwapGate
from ._det_annotation import DetAnnotation
from ._obs_annotation import CumulativeObservableAnnotation
from ._shift_coords_annotation import ShiftCoordsAnnotation
from ._stim_sampler import StimSampler
from ._stim_to_cirq import (
    MeasureAndOrResetGate,
    stim_circuit_to_cirq_circuit,
)
from ._sweep_pauli import SweepPauli
from ._two_qubit_asymmetric_depolarize import TwoQubitAsymmetricDepolarizingChannel
from ._i_error_gate import IErrorGate
from ._ii_error_gate import IIErrorGate
from ._ii_gate import IIGate

JSON_RESOLVERS_DICT = {
    "CumulativeObservableAnnotation": CumulativeObservableAnnotation,
    "DetAnnotation": DetAnnotation,
    "MeasureAndOrResetGate": MeasureAndOrResetGate,
    "ShiftCoordsAnnotation": ShiftCoordsAnnotation,
    "SweepPauli": SweepPauli,
    "TwoQubitAsymmetricDepolarizingChannel": TwoQubitAsymmetricDepolarizingChannel,
    "CXSwapGate": CXSwapGate,
    "CZSwapGate": CZSwapGate,
    "IIGate": IIGate,
    "IIErrorGate": IIErrorGate,
    "IErrorGate": IErrorGate,
}
JSON_RESOLVER = JSON_RESOLVERS_DICT.get



from typing import Any, Dict, List, Tuple

import cirq
import stim


@cirq.value_equality
class MeasureAndOrResetGate(cirq.Gate):
    """Handles explaining stim's MX/MY/MZ/MRX/MRY/MRZ/RX/RY/RZ gates to cirq."""

    def __init__(
        self,
        measure: bool,
        reset: bool,
        basis: str,
        invert_measure: bool,
        key: str,
        measure_flip_probability: float = 0,
    ):
        self.measure = measure
        self.reset = reset
        self.basis = basis
        self.invert_measure = invert_measure
        self.key = key
        self.measure_flip_probability = measure_flip_probability

    def _num_qubits_(self) -> int:
        return 1

    def _circuit_diagram_info_(self, args: cirq.CircuitDiagramInfoArgs) -> str:
        return str(self)

    def resolve(self, target: cirq.Qid) -> cirq.Operation:
        if (
            self.basis == 'Z'
            and self.measure
            and self.measure_flip_probability == 0
            and not self.reset
        ):
            return cirq.MeasurementGate(
                num_qubits=1, key=self.key, invert_mask=(1,) if self.invert_measure else ()
            ).on(target)

        return MeasureAndOrResetGate(
            measure=self.measure,
            reset=self.reset,
            basis=self.basis,
            invert_measure=self.invert_measure,
            key=self.key,
            measure_flip_probability=self.measure_flip_probability,
        ).on(target)

    def _value_equality_values_(self):
        return (
            self.measure,
            self.reset,
            self.basis,
            self.invert_measure,
            self.key,
            self.measure_flip_probability,
        )

    def _decompose_(self, qubits):
        (q,) = qubits
        if self.measure:
            if self.basis == 'X':
                yield cirq.H(q)
            elif self.basis == 'Y':
                yield cirq.X(q) ** 0.5
            if self.measure_flip_probability:
                raise NotImplementedError("Noisy measurement as a cirq operation.")
            else:
                yield cirq.measure(
                    q, key=self.key, invert_mask=(True,) if self.invert_measure else ()
                )
        if self.reset:
            yield cirq.ResetChannel().on(q)
        if self.measure or self.reset:
            if self.basis == 'X':
                yield cirq.H(q)
            elif self.basis == 'Y':
                yield cirq.X(q) ** -0.5

    def _stim_op_name(self) -> str:
        result = ''
        if self.measure:
            result += "M"
        if self.reset:
            result += "R"
        if self.basis != 'Z':
            result += self.basis
        return result

    def _stim_conversion_(self, *, edit_circuit: stim.Circuit, targets: List[int], tag: str, edit_measurement_key_lengths: List[Tuple[str, int]], **kwargs):
        if self.measure:
            edit_measurement_key_lengths.append((self.key, 1))
        if self.invert_measure:
            targets[0] = stim.target_inv(targets[0])
        if self.measure_flip_probability:
            edit_circuit.append_operation(
                self._stim_op_name(), targets, self.measure_flip_probability, tag=tag
            )
        else:
            edit_circuit.append_operation(self._stim_op_name(), targets, tag=tag)

    def __str__(self) -> str:
        result = self._stim_op_name()
        if self.invert_measure:
            result = "!" + result
        if self.measure:
            result += f"('{self.key}')"
        if self.measure_flip_probability:
            result += f"^{self.measure_flip_probability:%}"
        return result

    def __repr__(self):
        return (
            f'stimcirq.MeasureAndOrResetGate('
            f'measure={self.measure!r}, '
            f'reset={self.reset!r}, '
            f'basis={self.basis!r}, '
            f'invert_measure={self.invert_measure!r}, '
            f'key={self.key!r}, '
            f'measure_flip_probability={self.measure_flip_probability!r})'
        )

    @staticmethod
    def _json_namespace_() -> str:
        return ''

    def _json_dict_(self) -> Dict[str, Any]:
        return {
            'measure': self.measure,
            'reset': self.reset,
            'basis': self.basis,
            'invert_measure': self.invert_measure,
            'key': self.key,
            'measure_flip_probability': self.measure_flip_probability,
        }



from typing import Any, Dict, Iterable, List, Tuple

import cirq
import stim


@cirq.value_equality
class DetAnnotation(cirq.Operation):
    """Annotates that a particular combination of measurements is deterministic.

    Creates a DETECTOR operation when converting to a stim circuit.
    """

    def __init__(
        self,
        *,
        parity_keys: Iterable[str] = (),
        relative_keys: Iterable[int] = (),
        coordinate_metadata: Iterable[float] = (),
    ):
        """

        Args:
            parity_keys: The keys of some measurements with the property that their parity is always
                the same under noiseless execution of the circuit.
            relative_keys: Refers to measurements relative to this operation. For example,
                relative key -1 is the previous measurement. All entries must be negative.
            coordinate_metadata: An optional location for the detector. This has no effect on the
                function of the circuit, but can be used by plotting tools.
        """
        self.parity_keys = frozenset(parity_keys)
        self.relative_keys = frozenset(relative_keys)
        self.coordinate_metadata = tuple(coordinate_metadata)

    @property
    def qubits(self) -> Tuple[cirq.Qid, ...]:
        return ()

    def with_qubits(self, *new_qubits) -> 'DetAnnotation':
        return self

    def _value_equality_values_(self) -> Any:
        return self.parity_keys, self.relative_keys, self.coordinate_metadata

    def _circuit_diagram_info_(self, args: Any) -> str:
        items: List[str] = [repr(e) for e in sorted(self.parity_keys)]
        items += [f'rec[{e}]' for e in sorted(self.relative_keys)]
        k = ",".join(str(e) for e in items)
        return f"Det({k})"

    @staticmethod
    def _json_namespace_() -> str:
        return ''

    def _json_dict_(self) -> Dict[str, Any]:
        result = {
            'parity_keys': sorted(self.parity_keys),
            'coordinate_metadata': self.coordinate_metadata,
        }
        if self.relative_keys:
            result['relative_keys'] = sorted(self.relative_keys)
        return result

    def __repr__(self) -> str:
        return (
            f'stimcirq.DetAnnotation('
            f'parity_keys={sorted(self.parity_keys)}, '
            f'relative_keys={sorted(self.relative_keys)}, '
            f'coordinate_metadata={self.coordinate_metadata!r})'
        )

    def _decompose_(self):
        return []

    def _is_comment_(self) -> bool:
        return True

    def _stim_conversion_(
        self,
        *,
        edit_circuit: stim.Circuit,
        edit_measurement_key_lengths: List[Tuple[str, int]],
        tag: str,
        have_seen_loop: bool = False,
        **kwargs,
    ):
        # Ideally these references would all be resolved ahead of time, to avoid the redundant
        # linear search overhead and also to avoid the detectors and measurements being interleaved
        # instead of grouped (grouping measurements is helpful for stabilizer simulation). But that
        # didn't happen and this is the context we're called in and we're going to make it work.

        if have_seen_loop and self.parity_keys:
            raise NotImplementedError(
                "Measurement key conversion is not reliable when loops are present."
            )

        # Find indices of measurement record targets.
        remaining = set(self.parity_keys)
        rec_targets = [stim.target_rec(k) for k in sorted(self.relative_keys, reverse=True)]
        for offset in range(len(edit_measurement_key_lengths)):
            m_key, m_len = edit_measurement_key_lengths[-1 - offset]
            if m_len != 1:
                raise NotImplementedError(f"multi-qubit measurement {m_key!r}")
            if m_key in remaining:
                remaining.discard(m_key)
                rec_targets.append(stim.target_rec(-1 - offset))
                if not remaining:
                    break
        if remaining:
            raise ValueError(
                f"{self!r} was processed before measurements it referenced ({sorted(remaining)!r})."
                f" Make sure the referenced measurements keys are actually in the circuit, and come"
                f" in an earlier moment (or earlier in the same moment's operation order)."
            )

        edit_circuit.append("DETECTOR", rec_targets, self.coordinate_metadata, tag=tag)
